import datetime
import json
from os import path, makedirs
from os.path import join
from typing import Any, Dict, List
from pathlib import Path

import lmfit as lm
import matplotlib as mpl
import xarray as xr
from quantify_core.data.handling import (
    load_dataset_from_path,
)

REPO_PATH = path.abspath("").split("nuclear_spin_snv")[0] + "nuclear_spin_snv"
DATA_PATH = path.join(REPO_PATH, "data")
DATA_RESULTS_PATH = path.join(REPO_PATH, "data_results")
DATA_VALUES_PATH = path.join(REPO_PATH, "data_values")
FIGURE_PATH = path.join(REPO_PATH, "figures")
SUBFIGURE_PATH = path.join(REPO_PATH, "subfigures")


def _generate_date_time():
    time_stamp = datetime.datetime.now()
    # time_stamp gives microseconds by default
    (date_time, micro) = time_stamp.strftime("%Y%m%d-%H%M%S-.%f").split(".")
    # this ensures the string is formatted correctly as some systems return 0 for micro
    date_time = f"{date_time}{int(int(micro) / 1000):03d}-"
    return date_time

def _ensure_folders_exists(path_name):
    path = Path(path_name)
    directory = path.parent
    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)


def save_dataset(dataset, folder, name, timestamp=False):
    date_time = _generate_date_time() if timestamp else ""
    file_path = join(folder, date_time + name + ".hdf5")
    _ensure_folders_exists(file_path)
    # Invalid_netcdf is used to be able to save None and bools as attrs
    dataset.to_netcdf(file_path, engine="h5netcdf", invalid_netcdf=True)


def load_dataset(folder, name):
    return xr.load_dataset(join(folder, f"{name}.hdf5"), engine="h5netcdf")


def load_dataset_from_data(
    folder_name: str, dataset_name: str = "dataset_processed", results=False
) -> xr.Dataset:
    main_path = DATA_RESULTS_PATH if results else DATA_PATH
    return load_dataset(path.join(main_path, folder_name), dataset_name)


def load_quantities_of_interest_from_data(
    folder_name: str, qoi_name: str = "quantities_of_interest"
) -> Dict:
    # Load JSON file and return
    qoi_path = path.join(DATA_PATH, folder_name, f"{qoi_name}.json")
    with open(path.join(qoi_path), "r") as file:
        quantities_of_interest = json.load(file)
    return quantities_of_interest


def save_figure(
    fig: mpl.figure.Figure, figure_number: int, subfigure_name: str, **kwargs
) -> None:
    kwargs.setdefault("transparent", True)
    file_path = path.join(SUBFIGURE_PATH, f"figure{figure_number}", f"{subfigure_name}.svg")
    _ensure_folders_exists(file_path)
    fig.savefig(file_path, **kwargs)


def save_model_results(results: Dict[str, Any]) -> None:
    for name, r in results.items():
        _ensure_folders_exists(name)
        lm.model.save_modelresult(r, name)


def load_model_results(file_names: List[str], keys: List[str]) -> List[Any]:
    fit_results = {}
    for f, k in zip(file_names, keys):
        fit_results[k] = lm.model.load_modelresult(f)
    return fit_results


def save_txt(string, name):
    with open(join(DATA_VALUES_PATH, f"{name}.txt"), "w") as f:
        f.write(string)
