import operator

import lmfit as lm
import numpy as np
import xarray as xr
from scipy.stats import skew


def combine_params(*params_list):
    params = params_list[0]
    for params_new in params_list[1:]:
        param_new_list = list(dict(params_new).values())
        params.add_many(*param_new_list)
    return params


def exponential_envelope(x, amplitude=1, decay=1, exponent=1):
    """Return an exponential function.

    exponential(x, amplitude, decay) = amplitude * exp(-x/decay)

    """
    decay = lm.lineshapes.not_zero(decay)
    return amplitude * np.exp(-((x / decay) ** exponent))


class ExponentialEnvelopeModel(lm.Model):
    r"""A model based on an exponential decay function.

    The model has two Parameters: `amplitude` (:math:`A`), `decay`
    (:math:`\tau`), and `exponent` (:math:`n`) and is defined as:

    .. math::

        f(x; A, \tau) = A e^{-(x/\tau)^n}

    """

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super().__init__(exponential_envelope, **kwargs)

    def guess(self, data, x, **kwargs):
        """Estimate initial model parameter values from data."""
        max_e = max(data) / np.e
        difference = np.abs(data - max_e)
        decay = x[np.argmin(difference)]
        pars = self.make_params(amplitude=np.max(data), decay=decay, exponent=1)
        return lm.models.update_param_vals(pars, self.prefix, **kwargs)

    __init__.__doc__ = lm.models.COMMON_INIT_DOC
    guess.__doc__ = lm.models.COMMON_GUESS_DOC


class ExponentialEnvelopeConstantModel(lm.model.CompositeModel):
    __doc__ = "exponential envelope constant model" + lm.models.COMMON_INIT_DOC

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            left=ExponentialEnvelopeModel(),
            right=lm.models.ConstantModel(),
            op=operator.add,
            *args,
            **kwargs,
        )

    def guess(self, data, x, **kwargs):
        c = np.percentile(data, 10)
        data_cor = data - c
        params0 = self.components[0].guess(data_cor, x, **kwargs)
        params1 = self.components[1].make_params(c=c)
        params = combine_params(params0, params1)
        return lm.models.update_param_vals(params, self.prefix, **kwargs)


class SineConstantModel(lm.model.CompositeModel):
    __doc__ = "sine offset constant model" + lm.models.COMMON_INIT_DOC

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            left=lm.models.SineModel(),
            right=lm.models.ConstantModel(),
            op=operator.add,
            *args,
            **kwargs,
        )

    def guess(self, data, x, **kwargs):
        c = np.mean(data)
        data_cor = data - c
        params0 = self.components[0].guess(data_cor, x, **kwargs)
        params1 = self.components[1].make_params(const_c=c)
        params = combine_params(params0, params1)
        return lm.models.update_param_vals(params, self.prefix, **kwargs)


def sine_offset(x, amplitude=1.0, frequency=1.0, shift=0.0, offset=0.0):
    """Return a sinusoidal function.

    sine(x, amplitude, frequency, shift, offset) =
        amplitude * sin((x - offset)*frequency + shift)

    """
    return amplitude * np.sin((x - offset) * frequency + shift)


class SineOffsetModel(lm.Model):
    r"""A model based on a offsetted sinusoidal lineshape.

    The model has three Parameters: `amplitude`, `frequency`, `shift`, and `offset`.

    .. math::

        f(x; A, \phi, f, x0) = A \sin (f (x-x0) + \phi)

    where the parameter `amplitude` corresponds to :math:`A`, `frequency` to
    :math:`f`, `shift` to :math:`\phi`, and `offset` to :math:`x0`.
    All but x0 are constrained to be non-negative, and `shift` additionally to be smaller than :math:`2\pi`.

    """

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super().__init__(sine_offset, **kwargs)
        self._set_paramhints_prefix()

    def _set_paramhints_prefix(self):
        self.set_param_hint("amplitude", min=0)
        self.set_param_hint("frequency", min=0)
        self.set_param_hint("shift", min=-2 * np.pi - 1.0e-5, max=2 * np.pi + 1.0e-5)
        self.set_param_hint("offset", vary=False)

    def guess(self, data, x, **kwargs):
        """Estimate initial model parameter values from the FFT of the data."""
        data = data - data.mean()
        # assume uniform spacing
        frequencies = np.fft.fftfreq(len(x), abs(x[-1] - x[0]) / (len(x) - 1))
        fft = abs(np.fft.fft(data))
        argmax = abs(fft).argmax()
        amplitude = 2.0 * fft[argmax] / len(fft)
        frequency = 2 * np.pi * abs(frequencies[argmax])
        offset = x[0]
        # try shifts in the range [0, 2*pi) and take the one with best residual
        shift_guesses = np.linspace(0, 2 * np.pi, 11, endpoint=False)
        errors = [
            np.linalg.norm(
                self.eval(
                    x=x,
                    amplitude=amplitude,
                    frequency=frequency,
                    shift=shift_guess,
                    offset=offset,
                )
                - data
            )
            for shift_guess in shift_guesses
        ]
        shift = shift_guesses[np.argmin(errors)]
        pars = self.make_params(
            amplitude=amplitude, frequency=frequency, shift=shift, offset=offset
        )
        return lm.models.update_param_vals(pars, self.prefix, **kwargs)

    __init__.__doc__ = lm.models.COMMON_INIT_DOC
    guess.__doc__ = lm.models.COMMON_GUESS_DOC


class SineOffsetConstantModel(lm.model.CompositeModel):
    __doc__ = "sine offset constant model" + lm.models.COMMON_INIT_DOC

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            left=SineOffsetModel(),
            right=lm.models.ConstantModel(),
            op=operator.add,
            *args,
            **kwargs,
        )

    def guess(self, data, x, **kwargs):
        mask = ~np.isnan(data) * ~np.isnan(x)
        x = x[mask]
        data = data[mask]
        c = np.mean(data)
        data_cor = data - c
        params0 = self.components[0].guess(data_cor, x, **kwargs)
        params1 = self.components[1].make_params(c=c)
        params = combine_params(params0, params1)
        return lm.models.update_param_vals(params, self.prefix, **kwargs)


class DecayingSineConstantModel(lm.model.CompositeModel):
    __doc__ = "decaying sine constant model" + lm.models.COMMON_INIT_DOC

    def __init__(self, *args, **kws) -> None:
        super().__init__(
            left=lm.model.CompositeModel(
                left=lm.models.SineModel(),
                right=lm.models.ExponentialModel(prefix="exp_"),
                op=operator.mul,
            ),
            right=lm.models.ConstantModel(),
            op=operator.add,
            *args,
            **kws,
        )

    def guess(self, data, x, **kws):
        c = np.mean(data)
        data_cor = data - c
        params2 = self.components[2].make_params(c=c)
        params0 = self.components[0].guess(data_cor, x, **kws)
        params1 = self.components[1].make_params(
            exp_amplitude={"value": 1, "vary": False}
        )
        params = combine_params(params0, params1, params2)
        return lm.models.update_param_vals(params, self.prefix, **kws)


class DecayingEnvelopeSineConstantModel(lm.model.CompositeModel):
    __doc__ = "decaying sine constant model" + lm.models.COMMON_INIT_DOC

    def __init__(self, *args, **kws) -> None:
        super().__init__(
            left=lm.model.CompositeModel(
                left=lm.models.SineModel(),
                right=ExponentialEnvelopeModel(prefix="env_"),
                op=operator.mul,
            ),
            right=lm.models.ConstantModel(),
            op=operator.add,
            *args,
            **kws,
        )

    def guess(self, data, x, **kws):
        c = np.mean(data)
        data_cor = data - c
        params2 = self.components[2].make_params(c=c)
        params0 = self.components[0].guess(data_cor, x, **kws)
        params1 = self.components[1].make_params(
            env_amplitude={"value": 1, "vary": False}, env_decay={"value": 1, "min": 0}
        )
        params = combine_params(params0, params1, params2)
        return lm.models.update_param_vals(params, self.prefix, **kws)


class BeatingSineConstantModel(lm.model.CompositeModel):
    __doc__ = "decaying sine constant model" + lm.models.COMMON_INIT_DOC

    def __init__(self, *args, num_sines=2, **kws) -> None:
        model = lm.models.SineModel(prefix="sine0_")
        for i in range(1, num_sines):
            model = lm.model.CompositeModel(
                left=model,
                right=lm.models.SineModel(prefix=f"sine{i}_"),
                op=operator.add,
            )

        model = lm.model.CompositeModel(
            left=model,
            right=ExponentialEnvelopeModel(),
            op=operator.mul,
        )

        super().__init__(
            left=model,
            right=lm.models.ConstantModel(),
            op=operator.add,
            *args,
            **kws,
        )

    def guess(self, data, x, **kws):
        c = np.mean(data)
        params = self.make_params(c=c)
        return lm.models.update_param_vals(params, self.prefix, **kws)


class TwoVoigtConstantModel(lm.model.CompositeModel):
    __doc__ = "two voigt constant model" + lm.models.COMMON_INIT_DOC

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            left=lm.model.CompositeModel(
                left=lm.models.VoigtModel(prefix="voigt0_"),
                right=lm.models.VoigtModel(prefix="voigt1_"),
                op=operator.add,
            ),
            right=lm.models.ConstantModel(),
            op=operator.add,
            *args,
            **kwargs,
        )

    def guess(self, data, x, **kwargs):
        skewness = skew(data)
        if skewness > 0:
            c = np.percentile(data, 10)
        else:
            c = np.percentile(data, 90)
        params = self.make_params(c=c)
        return lm.models.update_param_vals(params, self.prefix, **kwargs)


class VoigtConstantModel(lm.model.CompositeModel):
    __doc__ = "voigt constant model" + lm.models.COMMON_INIT_DOC

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            left=lm.models.VoigtModel(),
            right=lm.models.ConstantModel(),
            op=operator.add,
            *args,
            **kwargs,
        )

    def guess(self, data, x, **kwargs):
        skewness = skew(data)
        if skewness > 0:
            c = np.percentile(data, 10)
        else:
            c = np.percentile(data, 90)
        paramsc = self.make_params(c=c)
        data_cor = data - c
        paramsv = self.components[0].guess(data_cor, x)
        params = combine_params(paramsc, paramsv)
        return lm.models.update_param_vals(params, self.prefix, **kwargs)


class DecayingCosineSquaredModel(lm.Model):
    def __init__(self, *args, **kwargs):
        def decaying_cosine_squared(x, amplitude, frequency, shift, decay):
            return amplitude * (
                (np.cos((x + shift) * 2 * np.pi * frequency / 2) ** 2 - 1 / 2)
                * np.exp(-(x - shift) / decay)
                + 1 / 2
            )

        super(DecayingCosineSquaredModel, self).__init__(
            decaying_cosine_squared, *args, **kwargs
        )

    def guess(self, data, x, **kwargs):
        # This method is copied from the sine build-in model
        data = data - data.mean()
        # assume uniform spacing
        frequencies = np.fft.fftfreq(len(x), abs(x[-1] - x[0]) / (len(x) - 1))
        fft = abs(np.fft.fft(data))
        argmax = abs(fft).argmax()
        amplitude = 2.0 * np.pi * fft[argmax] / len(fft) / np.sqrt(2)
        frequency = abs(frequencies[argmax])
        # try shifts in the range [0, 2*pi) and take the one with best residual
        shift_guesses = np.linspace(0, 2 * np.pi, 11, endpoint=False)
        errors = [
            np.linalg.norm(
                self.eval(
                    x=x,
                    amplitude=amplitude,
                    frequency=frequency,
                    shift=shift_guess,
                    decay=10e6,
                )
                - data
            )
            for shift_guess in shift_guesses
        ]
        shift = shift_guesses[np.argmin(errors)]

        # try shifts in the range [0, 2*pi) and take the one with best residual
        decay_guesses = np.geomspace(0.1, 1000, 5) * np.max(x)
        errors = [
            np.linalg.norm(
                self.eval(
                    x=x,
                    amplitude=amplitude,
                    frequency=frequency,
                    shift=shift,
                    decay=decay_guess,
                )
                - data
            )
            for decay_guess in decay_guesses
        ]
        decay = decay_guesses[np.argmin(errors)]

        params = self.make_params(
            amplitude=amplitude, frequency=frequency, shift=shift, decay=decay
        )
        return lm.models.update_param_vals(params, self.prefix, **kwargs)


def fit_saturation(x, max_counts, power_sat):
    return max_counts * x / (x + power_sat)


sat = lm.Model(fit_saturation)


def saturation_curve(x, amplitude=1, saturation=1):
    """Return an saturation curve.

    exponential(x, amplitude, saturation) = amplitude * x / (2*x + saturation)

    """
    return amplitude * x / (2 * x + saturation)


class SaturationCurveModel(lm.Model):
    r"""A model based on a saturation curve function.

    The model has two Parameters: `amplitude` (:math:`A`), and `saturation`
    (:math:`s`) and is defined as:

    .. math::

        f(x; A, s) = amplitude * x / (x + s)

    """

    def __init__(self, independent_vars=["x"], prefix="", nan_policy="raise", **kwargs):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super().__init__(saturation_curve, **kwargs)

    def guess(self, data, x, **kwargs):
        """Estimate initial model parameter values from data."""
        half_max = max(data) * 2 / 3
        difference = np.abs(data - half_max)
        saturation = x[np.argmin(difference)]
        pars = self.make_params(amplitude=np.max(data), saturation=saturation)
        return lm.models.update_param_vals(pars, self.prefix, **kwargs)

    __init__.__doc__ = lm.models.COMMON_INIT_DOC
    guess.__doc__ = lm.models.COMMON_GUESS_DOC


def _amplitude(rf_voltage, voltage_pi2):
    return np.cos(rf_voltage / voltage_pi2 * np.pi / 2 / 2) ** 2 * 2 - 1


def _f(rf_voltage, e_phase, voltage_pi2, phase_pi2, A):
    return (
        A
        * (
            -_amplitude(rf_voltage, voltage_pi2)
            * np.sin(
                e_phase / 180 * np.pi
                - phase_pi2 / 180 * np.pi * (rf_voltage / voltage_pi2)
            )
            # / 2 # needed when fitting p0
        )
        # + 0.5 # needed when fitting p0
    )


def ddrf_voltage_sweep(rf_voltage, e_phase, voltage_pi2, phase_pi2, A):
    ds = xr.Dataset()
    ds = ds.assign_coords(rf_voltage=rf_voltage)
    ds = ds.assign_coords(e_phase=e_phase)
    return xr.apply_ufunc(_f, ds.rf_voltage, ds.e_phase, voltage_pi2, phase_pi2, A)


class DDRFGateCalibrationModel(lm.Model):
    """
    Model for Gate Calibration.
    The model fits a cos^2 amplitude with respect to the rf drive voltage and a sine with respect to the readout phase
    """

    def __init__(
        self,
        independent_vars=["rf_voltage", "e_phase"],
        prefix="",
        nan_policy="raise",
        **kwargs,
    ):
        kwargs.update(
            {
                "prefix": prefix,
                "nan_policy": nan_policy,
                "independent_vars": independent_vars,
            }
        )
        super().__init__(ddrf_voltage_sweep, **kwargs)

    def guess(self, data, **kwargs):
        voltage_pi2 = float(data.std("e_phase").idxmin().values)
        phase_pi2 = 0
        A = data.max() - data.min()
        pars = self.make_params(voltage_pi2=voltage_pi2, phase_pi2=phase_pi2, A=A)
        return lm.models.update_param_vals(pars, self.prefix, **kwargs)

    __init__.__doc__ = lm.models.COMMON_INIT_DOC
    guess.__doc__ = lm.models.COMMON_GUESS_DOC
