import numpy as np
import xarray as xr
from snvc.simulations import hyperfine


def k(s0, s1, w0, w1, wl, Azx):
    return ((s1 - s0) * wl * Azx) / (w0 * w1)


def w0innerw1(s0, s1, w0, w1, wl, Azz, Azx):
    return (wl**2 + (s0 + s1) * Azz * wl + s0 * s1 * (Azz**2 + Azx**2)) / (w0 * w1)


def theta(s0, s1, w0, w1, wl, Azz, Azx, tau):
    w0innerw1_ = w0innerw1(s0, s1, w0, w1, wl, Azz, Azx)
    return np.arccos(
        np.cos(w0 * tau) * np.cos(w1 * tau)
        - w0innerw1_ * np.sin(w0 * tau) * np.sin(w1 * tau)
    )


# (1 - n0.n1)
def anti_alignment_n0n1(s0, s1, w0, w1, wl, Azz, Azx, tau):
    k_ = k(s0, s1, w0, w1, wl, Azx)
    theta_ = theta(s0, s1, w0, w1, wl, Azz, Azx, tau)
    return (
        2
        * k_**2
        * np.sin(w0 * tau / 2) ** 2
        * np.sin(w1 * tau / 2) ** 2
        / np.cos(theta_ / 2) ** 2
    )


def sigma_z(s0, s1, w0, w1, wl, Azz, Azx, tau, N):
    anti_alignment_n0n1_ = anti_alignment_n0n1(s0, s1, w0, w1, wl, Azz, Azx, tau)
    theta_ = theta_effective(s0, s1, w0, w1, wl, Azz, Azx, tau)
    return 1 - anti_alignment_n0n1_ * np.sin((N / 2) * theta_) ** 2


def sigma_z_A(s0, s1, wl, Azz, Azx, tau, N):
    w0, w1 = xr.apply_ufunc(
        hyperfine.calc_w, s0, s1, wl, Azz, Azx, output_core_dims=[[], []]
    )
    return sigma_z(s0, s1, w0, w1, wl, Azz, Azx, tau, N)


def sigma_z_w(s0, s1, w0, w1, wl, tau, N):
    Azz, Azx = xr.apply_ufunc(
        hyperfine.calc_A, s0, s1, w0, w1, wl, output_core_dims=[[], []]
    )
    return xr.apply_ufunc(sigma_z, s0, s1, w0, w1, wl, Azz, Azx, tau, N)


def bath(s0, s1, wl, tau, N, r_min, r_max, A_max=None, seed=0):
    Azz, Azx = xr.apply_ufunc(
        hyperfine.random_hyperfine,
        r_min,
        r_max,
        seed,
        vectorize=True,
        output_core_dims=[["c13"], ["c13"]],
    )
    z = xr.apply_ufunc(sigma_z_A, s0, s1, wl, Azz, Azx, tau, N)
    if A_max is not None:
        z_filtered = z.where(np.sqrt(Azz**2 + Azx**2) < A_max, 1)
    else:
        z_filtered = z
    bath = np.prod(z_filtered, axis=z_filtered.get_axis_num("c13"))
    return bath


def tau(w0, w1, p):
    return (2 * p + 1) * np.pi / (w0 + w1)


def effective_rotation_angle(angle_rad):
    angle_mod_rad = angle_rad % (2 * np.pi)
    if isinstance(angle_rad, float):
        if angle_mod_rad > np.pi / 2:
            angle_mod_rad -= np.pi
    elif isinstance(angle_rad, np.ndarray):
        angle_mod_rad = np.where(
            angle_mod_rad < np.pi / 2, angle_mod_rad, angle_mod_rad - np.pi
        )
    elif isinstance(angle_rad, xr.DataArray):
        angle_mod_rad = angle_mod_rad.where(
            angle_mod_rad < np.pi / 2, angle_mod_rad - np.pi
        )
    return np.abs(angle_mod_rad)


def theta_effective(s0, s1, w0, w1, wl, Azz, Azx, tau):
    return effective_rotation_angle(theta(s0, s1, w0, w1, wl, Azz, Azx, tau))


def N_gate(s0, s1, w0, w1, wl, Azz, Azx, tau):
    anti_alignment_n0n1_ = anti_alignment_n0n1(s0, s1, w0, w1, wl, Azz, Azx, tau)
    theta_ = theta_effective(s0, s1, w0, w1, wl, Azz, Azx, tau)
    N = 2 / theta_ * np.arcsin(1 / np.sqrt(anti_alignment_n0n1_))
    return N
