from functools import wraps

import numpy as np
import qutip as qt
import snvc.simulations.decoupling as dec
import snvc.xarray_mods as xrmod
import xarray as xr
from scipy.optimize import root_scalar


def handle_nan(func):
    """If np.nan gets passed as one of the arguments, return nan."""

    @wraps(func)
    def wrapped(*args, **kwargs):
        for arg in args:
            if isinstance(arg, float):
                if np.isnan(arg):
                    return np.nan
        for kwarg in kwargs.values():
            if isinstance(kwarg, float):
                if np.isnan(kwarg):
                    return np.nan
        return func(*args, **kwargs)

    return wrapped


X = qt.tensor(qt.operations.rx(np.pi), qt.qeye(2))
x = qt.tensor(qt.operations.rx(np.pi / 2), qt.qeye(2))
xm = qt.tensor(qt.operations.rx(-np.pi / 2), qt.qeye(2))
y = qt.tensor(qt.operations.ry(np.pi / 2), qt.qeye(2))
ym = qt.tensor(qt.operations.ry(-np.pi / 2), qt.qeye(2))
init = qt.tensor(qt.fock_dm(2, 0), qt.maximally_mixed_dm(2))


def H0(f0, f1):
    proj_0 = qt.basis(2, 0).proj()
    proj_1 = qt.basis(2, 1).proj()
    Iz = 0.5 * qt.sigmaz()
    return (qt.tensor(proj_0, f0 * Iz) + qt.tensor(proj_1, f1 * Iz)) * 2 * np.pi


Hx = qt.tensor(qt.qeye(2), 0.5 * qt.sigmax()) * 2 * np.pi
Hy = qt.tensor(qt.qeye(2), 0.5 * qt.sigmay()) * 2 * np.pi


def Hx_0_coef(t, args):
    return (
        args["Omega_0"]
        * np.cos((args["f_rf_0"]) * t * 2 * np.pi + args["phi_0"])
        * rf_envelope(
            t,
            args["t_delta"],
            args["f_rf_0"] + args["f_rot"],
            args["ddrf_roll_periods"],
            args["pi_pulse_duration"],
        )
    )


def Hy_0_coef(t, args):
    return (
        args["Omega_0"]
        * np.sin((args["f_rf_0"]) * t * 2 * np.pi + args["phi_0"])
        * rf_envelope(
            t,
            args["t_delta"],
            args["f_rf_0"] + args["f_rot"],
            args["ddrf_roll_periods"],
            args["pi_pulse_duration"],
        )
    )


def Hx_1_coef(t, args):
    return (
        args["Omega_1"]
        * np.cos((args["f_rf_1"]) * t * 2 * np.pi + args["phi_1"])
        * rf_envelope(
            t,
            args["t_delta"],
            args["f_rf_1"] + args["f_rot"],
            args["ddrf_roll_periods"],
            args["pi_pulse_duration"],
        )
    )


def Hy_1_coef(t, args):
    return (
        args["Omega_1"]
        * np.sin((args["f_rf_1"]) * t * 2 * np.pi + args["phi_1"])
        * rf_envelope(
            t,
            args["t_delta"],
            args["f_rf_1"] + args["f_rot"],
            args["ddrf_roll_periods"],
            args["pi_pulse_duration"],
        )
    )


def rf_envelope(t, t_delta, rf_frequency, ddrf_roll_periods, pi_pulse_duration):
    period = 1 / rf_frequency
    t_roll = period * ddrf_roll_periods
    t_start_roll_on = min(pi_pulse_duration / 2, t_delta / 2)
    t_stop_roll_on = min(t_start_roll_on + t_roll, t_delta / 2)
    t_start_roll_off = t_delta - t_stop_roll_on
    t_stop_roll_off = t_delta - t_start_roll_on

    if t < t_start_roll_on:
        return 0
    if t_start_roll_on <= t < t_stop_roll_on:
        return np.sin(((t - t_start_roll_on) / t_roll) * np.pi / 2) ** 2
    if t_stop_roll_on <= t < t_start_roll_off:
        return 1
    if t_start_roll_off <= t < t_stop_roll_off:
        t_delta = t_stop_roll_off - t_roll
        return np.sin(((t - t_delta) / t_roll) * np.pi / 2 + np.pi / 2) ** 2
    if t_stop_roll_off < t:
        return 0


@handle_nan
def c_pi2_x(
    dm,
    f0,
    f1,
    rf_f0,
    rf_f1,
    f0_cal,
    f1_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    if Omega1 == 0:
        f_rot = rf_f0
    elif Omega0 == 0:
        f_rot = rf_f1
    else:
        f_rot = (rf_f0 + rf_f1) / 2
    f0 -= f_rot
    f1 -= f_rot
    rf_f0 -= f_rot
    rf_f1 -= f_rot
    f0_cal -= f_rot
    f1_cal -= f_rot

    dec_dict = dec.phase_dd(
        N, total_time, decoupling, f0_cal, f1_cal, phase_difference=np.pi
    )

    for i, (t_delta, rf_phase0, rf_phase1) in enumerate(
        zip(dec_dict["time_delta"], dec_dict["rf_phase0"], dec_dict["rf_phase1"])
    ):
        H_tot = [
            H0(f0=f0, f1=f1),
            [Hx, Hx_0_coef],
            [Hy, Hy_0_coef],
            [Hx, Hx_1_coef],
            [Hy, Hy_1_coef],
        ]
        args = {
            "f_rf_0": rf_f0,
            "f_rf_1": rf_f1,
            "Omega_0": Omega0,
            "phi_0": rf_phase0,
            "Omega_1": Omega1,
            "phi_1": rf_phase1,
            "t_delta": t_delta,
            "f_rot": f_rot,
            "ddrf_roll_periods": ddrf_roll_periods,
            "pi_pulse_duration": pi_pulse_duration,
        }
        result = qt.mesolve(
            H_tot,
            dm,
            [0, t_delta],
            options=qt.Options(store_final_state=True, nsteps=5000),
            args=args,
        )
        dm = result.final_state
        if i != len(dec_dict["time_delta"]) - 1:
            dm = X * dm * X.dag()
    return dm


def spectroscopy(
    f0,
    f1,
    rf_f0,
    rf_f1,
    f0_cal,
    f1_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    dm = y * init * y.dag()
    dm = c_pi2_x(
        dm,
        f0,
        f1,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )
    dm = ym * dm * ym.dag()
    return dm


def mbi(
    dm,
    f0,
    f1,
    rf_f0,
    rf_f1,
    f0_cal,
    f1_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    dm = y * dm * y.dag()
    dm = c_pi2_x(
        dm,
        f0,
        f1,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )
    dm = x * dm * x.dag()
    return dm


def sz(dm):
    dm_proj = dm.ptrace([0])
    return 2 * abs((qt.basis(2, 0).dag() * dm_proj * qt.basis(2, 0)).full()[0][0]) - 1


def ramsey(
    t_wait,
    f0,
    f1,
    rf_f0,
    rf_f1,
    f0_cal,
    f1_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    dm = init
    dm = mbi(
        dm,
        f0,
        f1,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )
    proj_0 = qt.tensor(qt.basis(2, 0).proj(), qt.qeye(2))
    dm = proj_0 * dm * proj_0.dag()
    result = qt.mesolve(
        H0(f0, f1), dm, [0, t_wait], options=qt.Options(store_final_state=True)
    )
    dm = result.final_state
    dm = mbi(
        dm,
        f0,
        f1,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )
    return sz(dm)


def resonant(rf_f0, f_avg_cal):
    rf_f1 = 2 * f_avg_cal - rf_f0
    f0_cal = rf_f0
    f1_cal = rf_f1
    return f0_cal, f1_cal, rf_f0, rf_f1


def snv(A_par, A_perp, f_larmor):
    f0 = np.sqrt((f_larmor + 0.5 * A_par) ** 2 + (0.5 * A_perp) ** 2)
    f1 = np.sqrt((f_larmor - 0.5 * A_par) ** 2 + (0.5 * A_perp) ** 2)
    return f0, f1


def contrast_spectrum(
    f0,
    f1,
    rf_f0,
    rf_f1,
    f0_cal,
    f1_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    dm = spectroscopy(
        f0,
        f1,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )
    return sz(dm)


def contrast_resonant(
    f_rf,
    f_avg_cal,
    Omega,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    f0_cal, f1_cal, rf_f0, rf_f1 = resonant(f_rf, f_avg_cal)
    Omega0 = Omega
    Omega1 = 0
    return contrast_spectrum(
        f0_cal,
        f1_cal,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )


def contrast_snv(
    A_par,
    A_perp,
    f_larmor,
    f_rf,
    f_avg_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    f0, f1 = snv(A_par, A_perp, f_larmor)
    f0_cal, f1_cal, rf_f0, rf_f1 = resonant(f_rf, f_avg_cal)
    return contrast_spectrum(
        f0,
        f1,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )


def contrast_spin(
    f0,
    f1,
    f_rf,
    f_avg_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    f0_cal, f1_cal, rf_f0, rf_f1 = resonant(f_rf, f_avg_cal)
    return contrast_spectrum(
        f0,
        f1,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )


def ramsey_snv(
    t_wait,
    A_par,
    A_perp,
    f_larmor,
    f_rf,
    f_avg_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
):
    f0, f1 = snv(A_par, A_perp, f_larmor)
    f0_cal, f1_cal, rf_f0, rf_f1 = resonant(f_rf, f_avg_cal)
    return ramsey(
        t_wait,
        f0,
        f1,
        rf_f0,
        rf_f1,
        f0_cal,
        f1_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )


def A_par_density(A_par):
    mu0 = np.pi * 4e-7
    g_e = 2.8031e6 * 1e4
    g_c = 1.0705e3 * 1e4
    alpha = mu0 * g_c * g_e / (4 * np.pi)
    c13_dens = 0.01109
    A_par_dens = (
        1e-3 * np.pi**2 * alpha * c13_dens / (A_par * 2 * np.pi) ** 2
    )  # spins / Hz
    return A_par_dens


def bath(
    f_larmor,
    f_rf,
    f_avg_cal,
    Omega0,
    Omega1,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
    A_par_max,
    A_par_step,
):
    pars = xr.Dataset()
    pars["A_par"] = np.arange(A_par_step, A_par_max, A_par_step)
    pars["A_perp"] = 0
    z = xrmod.apply_ufunc(
        contrast_snv,
        pars.A_par,
        pars.A_perp,
        f_larmor,
        f_rf,
        f_avg_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
        vectorize=True,
        multiprocessing=True,
    )

    A_par_dens = np.clip(A_par_density(pars.A_par), a_min=0, a_max=10)
    A_par_delta = pars.A_par[1] - pars.A_par[0]
    FACTOR_FOR_NEGATIVE_A_PAR = 2
    power = A_par_dens * A_par_delta * FACTOR_FOR_NEGATIVE_A_PAR
    weigthed = z**power
    ds_bath = np.prod(weigthed, axis=weigthed.get_axis_num("A_par"))
    mask = ~weigthed.isnull().any("A_par")
    ds_bath = ds_bath.where(mask)
    return ds_bath


def bath_gate(
    f_larmor,
    f_rf,
    f_avg_cal,
    N,
    total_time,
    decoupling,
    ddrf_roll_periods,
    pi_pulse_duration,
    A_par_max,
    A_par_step,
):
    Omega_needed = xrmod.apply_ufunc(
        find_Omega,
        f_rf,
        f_avg_cal,
        total_time,
        N,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
        vectorize=True,
        multiprocessing=True,
    )
    Omega0 = Omega_needed
    Omega1 = 0
    bath_array = bath(
        f_larmor,
        f_rf,
        f_avg_cal,
        Omega0,
        Omega1,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
        A_par_max,
        A_par_step,
    )
    return bath_array, Omega_needed


def electron_coherence(N, total_time, exp_N=0.47, A=132e-6, exp_t=2):
    decay = A * N**exp_N
    return np.exp(-((total_time / decay) ** exp_t))


def effective_Omega_factor(f0, f1, f_rf, time):
    Delta0 = f0 - f_rf
    Delta1 = f1 - f_rf
    return abs(np.sinc(Delta0 * time * 2 * np.pi) - np.sinc(Delta1 * time * 2 * np.pi))


def calculate_Omega_cpmg(
    f_rf, f_avg_cal, N, total_time, ddrf_roll_periods, pi_pulse_duration
):
    f0_cal, f1_cal, rf_f0, rf_f1 = resonant(f_rf, f_avg_cal)
    period = 1 / f_rf
    tau = total_time / (2 * N)
    INTEGRATION_CORRECTION = 0.5
    ROLL_ON_AND_ROLL_OFF = 2
    full_two_tau_pulse = 2 * tau - 2 * (
        pi_pulse_duration + period * ddrf_roll_periods * ROLL_ON_AND_ROLL_OFF
    )
    two_tau_pulse = (
        2 * tau
        - pi_pulse_duration
        - (period * ddrf_roll_periods * INTEGRATION_CORRECTION * ROLL_ON_AND_ROLL_OFF)
    )
    driving_time = two_tau_pulse * N
    if isinstance(driving_time, float):
        if driving_time < 0:
            return np.nan
    else:
        driving_time[full_two_tau_pulse <= 0] = np.nan
    uncorrected_Omega = 1 / (driving_time * 2)
    corrected_Omega = (
        uncorrected_Omega
        * 1
        / effective_Omega_factor(f0_cal, f1_cal, f_rf, two_tau_pulse)
    )
    if isinstance(driving_time, float):
        if corrected_Omega > 10e3:
            return np.nan
    else:
        corrected_Omega[corrected_Omega > 10e3] = np.nan
    return corrected_Omega


def find_Omega(
    f_rf, f_avg_cal, total_time, N, decoupling, ddrf_roll_periods, pi_pulse_duration
):
    guess = calculate_Omega_cpmg(
        f_rf, f_avg_cal, N, total_time, ddrf_roll_periods, pi_pulse_duration
    )
    if np.isnan(guess):
        return np.nan

    def func(Omega):
        return contrast_resonant(
            f_rf,
            f_avg_cal,
            Omega,
            N,
            total_time,
            decoupling,
            ddrf_roll_periods,
            pi_pulse_duration,
        )

    try:
        result = root_scalar(
            func, x0=0.8 * guess, x1=1.2 * guess, xtol=0.01, maxiter=20
        )
    except Exception:
        return np.nan
    if not result.converged:
        return np.nan
    return abs(result.root)


def check_gate(
    f_rf, f_avg_cal, N, total_time, decoupling, ddrf_roll_periods, pi_pulse_duration
):
    Omega = find_Omega(
        f_rf, f_avg_cal, total_time, N, decoupling, ddrf_roll_periods, pi_pulse_duration
    )
    return contrast_resonant(
        f_rf,
        f_avg_cal,
        Omega,
        N,
        total_time,
        decoupling,
        ddrf_roll_periods,
        pi_pulse_duration,
    )
