import numpy as np
import xarray as xr

mu0 = np.pi * 4e-7
h_bar = 1.0545718e-34  # J
gamma_e = 28.031e9 * 2 * np.pi  # 1/T
gamma_c = 10.705e6 * 2 * np.pi  # 1/T


def hyperfine(r, theta):
    Azx_J = (
        -mu0
        / (4 * np.pi * r**3)
        * (gamma_e * gamma_c * h_bar**2)
        * (3 * np.cos(theta) * np.sin(theta))
    )
    Azz_J = (
        -mu0
        / (4 * np.pi * r**3)
        * (gamma_e * gamma_c * h_bar**2)
        * (3 * np.cos(theta) ** 2 - 1)
    )
    Azx = Azx_J / h_bar
    Azz = Azz_J / h_bar
    return Azz, Azx


def random_r(size, r_min, r_max, seed=0):
    rng = np.random.default_rng(seed)
    samp = rng.uniform(size=size)
    # sample randomly r
    return ((r_max**3 - r_min**3) * samp + r_min**3) ** (1 / 3)


def random_theta(size, seed=0):
    rng = np.random.default_rng(seed + 1_000_000)
    samp = rng.uniform(size=size)  # *np.pi
    # sample randomly theta
    return np.arccos(1 - 2 * samp)


def sample_volume(r_min, r_max):
    return 4 / 3 * np.pi * (r_max**3 - r_min**3)


def number_c13(volume):
    c13_dens = 0.01109
    atom_dens = 1.764e29  # atoms / m3
    N = volume * atom_dens
    samplesize = int(N * c13_dens)
    return samplesize


def random_hyperfine(r_min, r_max, seed=0):
    samplesize = number_c13(sample_volume(r_min, r_max))
    r_array = random_r(samplesize, r_min, r_max, seed)
    r = xr.DataArray(r_array, dims="c13")
    theta_array = random_theta(samplesize, seed)
    theta = xr.DataArray(theta_array, dims="c13")
    Azz, Azx = xr.apply_ufunc(hyperfine, r, theta, output_core_dims=[[], []])
    return Azz, Azx


def calc_A(s0, s1, w0: float, w1: float, wl: float):
    """Calculate coupling parameters for a given set of evolution frequencies in a Ramsey experiment
    with a spin-1/2 system. Formula assumes A/f_lamor_C13 to be small.

    Attributes:
    f_0: Hz, Evolution frequency with eletron in state 0 (state after MBI projection)
    f_1: Hz, Evolution frequency with electron in state 1 (state oppoiste to MBI projection)
    B: T, magnetic field
    Formula from https://journals.aps.org/prx/pdf/10.1103/PhysRevX.12.011056

    Returns:
    Tuple of A_parallel, A_perpendicular in Hz
    """

    # f_larmor = GYROMAGNETIC_RATIO_C13 * B
    Azz = (s0**2 * (wl**2 - w1**2) - s1**2 * (wl**2 - w0**2)) / (
        (s1 - s0) * 2 * s0 * s1 * wl
    )
    Azx = np.sqrt(w1**2 - (wl + s1 * Azz) ** 2) / s1
    return Azz, Azx


def calc_w(s0, s1, wl, Azz, Azx):
    w0 = np.sqrt((wl + s0 * Azz) ** 2 + (s0 * Azx) ** 2)
    w1 = np.sqrt((wl + s1 * Azz) ** 2 + (s1 * Azx) ** 2)
    return w0, w1
